@@ -3,9 +3,10 @@
  * Copyright (C) 2007  Red Hat
  * see file 'COPYING' for use and warranty information
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; version 2 only
+ * This program is free software you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -13,8 +14,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #define _GNU_SOURCE
@@ -27,19 +27,20 @@
 #include <stdio.h>
 #include <stdarg.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <string.h>
 #include <errno.h>
 #include <time.h>
-#define KRB5_PRIVATE 1
 #include <krb5.h>
-#ifdef WITH_MOZLDAP
-#include <mozldap/ldap.h>
-#else
 #include <ldap.h>
-#endif
 #include <sasl/sasl.h>
 #include <popt.h>
 
+#include "config.h"
+
+#include "ipa_krb5.h"
+#include "ipa-client-common.h"
+
 /* Salt types */
 #define NO_SALT                        -1
 #define KRB5_KDB_SALTTYPE_NORMAL        0
@@ -49,8 +50,8 @@
 #define KRB5_KDB_SALTTYPE_SPECIAL       4
 #define KRB5_KDB_SALTTYPE_AFS3          5
 
-#define KEYTAB_SET_OID "2.16.840.1.113730.3.8.3.1"
-#define KEYTAB_RET_OID "2.16.840.1.113730.3.8.3.2"
+#define KEYTAB_SET_OID "2.16.840.1.113730.3.8.10.1"
+#define KEYTAB_RET_OID "2.16.840.1.113730.3.8.10.2"
 
 struct krb_key_salt {
     krb5_enctype enctype;
@@ -69,15 +70,33 @@ static int ldap_sasl_interact(LDAP *ld, unsigned flags, void *priv_data, void *s
 	sasl_interact_t *in = NULL;
 	int ret = LDAP_OTHER;
 	krb5_principal princ = (krb5_principal)priv_data;
+	krb5_context krbctx;
+	char *outname = NULL;
+	krb5_error_code krberr;
 
 	if (!ld) return LDAP_PARAM_ERROR;
 
 	for (in = sit; in && in->id != SASL_CB_LIST_END; in++) {
 		switch(in->id) {
 		case SASL_CB_USER:
-			in->result = princ->data[0].data;
-			in->len = princ->data[0].length;
+			krberr = krb5_init_context(&krbctx);
+
+			if (krberr) {
+				fprintf(stderr, _("Kerberos context initialization failed\n"));
+				in->result = NULL;
+				in->len = 0;
+				ret = LDAP_LOCAL_ERROR;
+				break;
+			}
+
+			krb5_unparse_name(krbctx, princ, &outname);
+
+			in->result = outname;
+			in->len = strlen(outname);
 			ret = LDAP_SUCCESS;
+
+			krb5_free_context(krbctx);
+
 			break;
 		case SASL_CB_GETREALM:
 			in->result = princ->realm.data;
@@ -90,7 +109,7 @@ static int ldap_sasl_interact(LDAP *ld, unsigned flags, void *priv_data, void *s
 			ret = LDAP_OTHER;
 		}
 	}
-        return ret;
+	return ret;
 }
 
 static void free_keys_contents(krb5_context krbctx, struct keys_container *keys)
@@ -125,7 +144,7 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
 
         krberr = krb5_get_permitted_enctypes(krbctx, &ktypes);
         if (krberr) {
-            fprintf(stderr, "No system preferred enctypes ?!\n");
+            fprintf(stderr, _("No system preferred enctypes ?!\n"));
             return 0;
         }
 
@@ -133,7 +152,7 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
 
         ksdata = calloc(n + 1, sizeof(struct krb_key_salt));
         if (NULL == ksdata) {
-            fprintf(stderr, "Out of memory!?\n");
+            fprintf(stderr, _("Out of memory!?\n"));
             return 0;
         }
 
@@ -142,7 +161,7 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
             ksdata[i].salttype = KRB5_KDB_SALTTYPE_NORMAL;
         }
 
-        krb5_free_ktypes(krbctx, ktypes);
+        ipa_krb5_free_ktypes(krbctx, ktypes);
 
         nkeys = i;
 
@@ -151,7 +170,7 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
 
         t = tmp = strdup(str);
         if (!tmp) {
-            fprintf(stderr, "Out of memory\n");
+            fprintf(stderr, _("Out of memory\n"));
             return 0;
         }
 
@@ -166,7 +185,7 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
         /* at the end we will have at most n entries + 1 terminating */
         ksdata = calloc(n + 1, sizeof(struct krb_key_salt));
         if (!ksdata) {
-            fprintf(stderr, "Out of memory\n");
+            fprintf(stderr, _("Out of memory\n"));
             return 0;
         }
 
@@ -181,7 +200,7 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
             krberr = krb5_string_to_enctype(t, &ksdata[j].enctype);
             if (krberr != 0) {
                 fprintf(stderr,
-                        "Warning unrecognized encryption type: [%s]\n", t);
+                        _("Warning unrecognized encryption type: [%s]\n"), t);
                 t = p+1;
                 continue;
             }
@@ -195,7 +214,8 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
 
             krberr = krb5_string_to_salttype(q, &ksdata[j].salttype);
             if (krberr != 0) {
-                fprintf(stderr, "Warning unrecognized salt type: [%s]\n", q);
+                fprintf(stderr,
+                        _("Warning unrecognized salt type: [%s]\n"), q);
                 continue;
             }
 
@@ -212,7 +232,7 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
      * MIT code do anyway */
 
     for (i = 0, n = 0; i < nkeys; i++ ) {
-        int similar = 0;
+        krb5_boolean similar = 0;
 
         for (j = 0; j < i; j++) {
             krberr = krb5_c_enctype_compare(krbctx,
@@ -221,7 +241,8 @@ static int prep_ksdata(krb5_context krbctx, const char *str,
                                             &similar);
             if (krberr) {
                 free_keys_contents(krbctx, keys);
-                fprintf(stderr, "Enctype comparison failed!\n");
+                free(ksdata);
+                fprintf(stderr, _("Enctype comparison failed!\n"));
                 return 0;
             }
             if (similar &&
@@ -258,7 +279,7 @@ static int create_keys(krb5_context krbctx,
     krb5_error_code krberr;
     krb5_data key_password;
     krb5_data *realm;
-    int i, j, nkeys;
+    int i, nkeys;
     int ret;
 
     ret = prep_ksdata(krbctx, enctypes_string, keys);
@@ -283,7 +304,7 @@ static int create_keys(krb5_context krbctx,
                                             ksdata[i].enctype,
                                             &ksdata[i].key);
             if (krberr) {
-                fprintf(stderr, "Failed to create random key!\n");
+                fprintf(stderr, _("Failed to create random key!\n"));
                 return 0;
             }
             /* set the salt to NO_SALT as the key was random */
@@ -296,14 +317,14 @@ static int create_keys(krb5_context krbctx,
         case KRB5_KDB_SALTTYPE_ONLYREALM:
             krberr = krb5_copy_data(krbctx, realm, &salt);
             if (krberr) {
-                fprintf(stderr, "Failed to create key!\n");
+                fprintf(stderr, _("Failed to create key!\n"));
                 return 0;
             }
 
             ksdata[i].salt.length = salt->length;
             ksdata[i].salt.data = malloc(salt->length);
             if (!ksdata[i].salt.data) {
-                fprintf(stderr, "Out of memory!\n");
+                fprintf(stderr, _("Out of memory!\n"));
                 return 0;
             }
             memcpy(ksdata[i].salt.data, salt->data, salt->length);
@@ -311,9 +332,9 @@ static int create_keys(krb5_context krbctx,
             break;
 
         case KRB5_KDB_SALTTYPE_NOREALM:
-            krberr = krb5_principal2salt_norealm(krbctx, princ, &ksdata[i].salt);
+            krberr = ipa_krb5_principal2salt_norealm(krbctx, princ, &ksdata[i].salt);
             if (krberr) {
-                fprintf(stderr, "Failed to create key!\n");
+                fprintf(stderr, _("Failed to create key!\n"));
                 return 0;
             }
             break;
@@ -321,7 +342,7 @@ static int create_keys(krb5_context krbctx,
         case KRB5_KDB_SALTTYPE_NORMAL:
             krberr = krb5_principal2salt(krbctx, princ, &ksdata[i].salt);
             if (krberr) {
-                fprintf(stderr, "Failed to create key!\n");
+                fprintf(stderr, _("Failed to create key!\n"));
                 return 0;
             }
             break;
@@ -336,7 +357,7 @@ static int create_keys(krb5_context krbctx,
              */
             ksdata[i].salt.data = (char *)malloc(realm->length + 1);
             if (NULL == ksdata[i].salt.data) {
-                fprintf(stderr, "Out of memory!\n");
+                fprintf(stderr, _("Out of memory!\n"));
                 return 0;
             }
             memcpy((char *)ksdata[i].salt.data,
@@ -347,7 +368,7 @@ static int create_keys(krb5_context krbctx,
             break;
 
         default:
-            fprintf(stderr, "Bad or unsupported salt type (%d)!\n",
+            fprintf(stderr, _("Bad or unsupported salt type (%d)!\n"),
                 ksdata[i].salttype);
             return 0;
         }
@@ -358,7 +379,7 @@ static int create_keys(krb5_context krbctx,
                                       &ksdata[i].salt,
                                       &ksdata[i].key);
         if (krberr) {
-            fprintf(stderr, "Failed to create key!\n");
+            fprintf(stderr, _("Failed to create key!\n"));
             return 0;
         }
 
@@ -457,7 +478,8 @@ int filter_keys(krb5_context krbctx, struct keys_container *keys,
             krb5_free_keyblock_contents(krbctx, &ksdata[i].key);
             krb5_free_data_contents(krbctx, &ksdata[i].salt);
             for (j = i; j < n-1; j++) {
-                keys[j] = keys[j + 1];
+                ksdata[j] = ksdata[j + 1];
+                enctypes[j] = enctypes[j + 1];
             }
             n--;
             /* new key has been moved to this position, make sure
@@ -467,7 +489,7 @@ int filter_keys(krb5_context krbctx, struct keys_container *keys,
     }
 
     if (n == 0) {
-        fprintf(stderr, "No keys accepted by KDC\n");
+        fprintf(stderr, _("No keys accepted by KDC\n"));
         return 0;
     }
 
@@ -475,10 +497,29 @@ int filter_keys(krb5_context krbctx, struct keys_container *keys,
     return n;
 }
 
+static int ipa_ldap_init(LDAP ** ld, const char * scheme, const char * servername, const int  port)
+{
+	char* url = NULL;
+	int  url_len = snprintf(url,0,"%s://%s:%d",scheme,servername,port) +1;
+       
+	url = (char *)malloc (url_len);
+	if (!url){
+		fprintf(stderr, _("Out of memory \n"));
+		return LDAP_NO_MEMORY;
+	}
+	sprintf(url,"%s://%s:%d",scheme,servername,port);
+	int rc = ldap_initialize(ld, url);
+
+	free(url);
+	return rc;
+}
+
 static int ldap_set_keytab(krb5_context krbctx,
 			   const char *servername,
 			   const char *principal_name,
 			   krb5_principal princ,
+			   const char *binddn,
+			   const char *bindpw,
 			   struct keys_container *keys)
 {
 	int version;
@@ -501,39 +542,77 @@ static int ldap_set_keytab(krb5_context krbctx,
 	/* cant' return more than nkeys, sometimes less */
 	encs = calloc(keys->nkeys + 1, sizeof(ber_int_t));
 	if (!encs) {
-		fprintf(stderr, "Out of Memory!\n");
+		fprintf(stderr, _("Out of Memory!\n"));
 		return 0;
 	}
 
 	/* build password change control */
 	control = create_key_control(keys, principal_name);
 	if (!control) {
-		fprintf(stderr, "Failed to create control!\n");
+		fprintf(stderr, _("Failed to create control!\n"));
 		goto error_out;
 	}
 
 	/* TODO: support referrals ? */
-	ld = ldap_init(servername, 389);
+	if (binddn) {
+		int ssl = LDAP_OPT_X_TLS_HARD;;
+		if (ldap_set_option(NULL, LDAP_OPT_X_TLS_CACERTFILE, "/etc/ipa/ca.crt") != LDAP_OPT_SUCCESS) {
+			goto error_out;
+		}
+ 
+		if ( ipa_ldap_init(&ld, "ldaps",servername, 636) != LDAP_SUCCESS){
+		  goto error_out;
+		}
+		if (ldap_set_option(ld, LDAP_OPT_X_TLS, &ssl) != LDAP_OPT_SUCCESS) {
+			goto error_out;
+		}
+	} else {
+		if (ipa_ldap_init(&ld, "ldap",servername, 389) != LDAP_SUCCESS){
+			goto error_out;			
+		}
+	}
+
 	if(ld == NULL) {
-		fprintf(stderr, "Unable to initialize ldap library!\n");
+		fprintf(stderr, _("Unable to initialize ldap library!\n"));
 		goto error_out;
 	}
 
+        /* Don't do DNS canonicalization */
+	ret = ldap_set_option(ld, LDAP_OPT_X_SASL_NOCANON, LDAP_OPT_ON);
+	if (ret != LDAP_SUCCESS) {
+	    fprintf(stderr, _("Unable to set LDAP_OPT_X_SASL_NOCANON\n"));
+	    goto error_out;
+	}
+
 	version = LDAP_VERSION3;
 	ret = ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &version);
         if (ret != LDAP_SUCCESS) {
-		fprintf(stderr, "Unable to set ldap options!\n");
+		fprintf(stderr, _("Unable to set ldap options!\n"));
 		goto error_out;
 	}
 
-	ret = ldap_sasl_interactive_bind_s(ld,
-					   NULL, "GSSAPI",
-					   NULL, NULL,
-					   LDAP_SASL_QUIET,
-					   ldap_sasl_interact, princ);
-	if (ret != LDAP_SUCCESS) {
-		fprintf(stderr, "SASL Bind failed!\n");
-		goto error_out;
+	if (binddn) {
+                struct berval bv;
+
+                bv.bv_val = discard_const(bindpw);
+                bv.bv_len = strlen(bindpw);
+
+                ret = ldap_sasl_bind_s(ld, binddn, LDAP_SASL_SIMPLE, &bv,
+                                       NULL, NULL, NULL);
+		if (ret != LDAP_SUCCESS) {
+			fprintf(stderr, _("Simple bind failed\n"));
+			goto error_out;
+		}
+	} else {
+		ret = ldap_sasl_interactive_bind_s(ld,
+						   NULL, "GSSAPI",
+						   NULL, NULL,
+						   LDAP_SASL_QUIET,
+						   ldap_sasl_interact, princ);
+		if (ret != LDAP_SUCCESS) {
+			fprintf(stderr, _("SASL Bind failed!\n"));
+			goto error_out;
+		}
 	}
 
 	/* find base dn */
@@ -547,7 +626,8 @@ static int ldap_set_keytab(krb5_context krbctx,
 					control, NULL, NULL,
 					&msgid);
 	if (ret != LDAP_SUCCESS) {
-		fprintf(stderr, "Operation failed! %s\n", ldap_err2string(ret));
+		fprintf(stderr, _("Operation failed! %s\n"),
+                                ldap_err2string(ret));
 		goto error_out;
 	}
 
@@ -559,24 +639,27 @@ static int ldap_set_keytab(krb5_context krbctx,
 
 	ret = ldap_result(ld, msgid, 1, &tv, &res);
 	if (ret == -1) {
-		fprintf(stderr, "Operation failed! %s\n", ldap_err2string(ret));
+		fprintf(stderr, _("Operation failed! %s\n"),
+                                ldap_err2string(ret));
 		goto error_out;
 	}
 
 	ret = ldap_parse_extended_result(ld, res, &retoid, &retdata, 0);
 	if(ret != LDAP_SUCCESS) {
-		fprintf(stderr, "Operation failed! %s\n", ldap_err2string(ret));
+		fprintf(stderr, _("Operation failed! %s\n"),
+                                ldap_err2string(ret));
 		goto error_out;
 	}
 
 	ret = ldap_parse_result(ld, res, &rc, NULL, &err, NULL, &srvctrl, 0);
         if(ret != LDAP_SUCCESS || rc != LDAP_SUCCESS) {
-		fprintf(stderr, "Operation failed! %s\n", err?err:ldap_err2string(ret));
+		fprintf(stderr, _("Operation failed! %s\n"),
+                                err ? err : ldap_err2string(ret));
 		goto error_out;
         }
 
 	if (!srvctrl) {
-		fprintf(stderr, "Missing reply control!\n");
+		fprintf(stderr, _("Missing reply control!\n"));
 		goto error_out;
 	}
 
@@ -586,14 +669,14 @@ static int ldap_set_keytab(krb5_context krbctx,
 		}
 	}
 	if (!pprc) {
-		fprintf(stderr, "Missing reply control!\n");
+		fprintf(stderr, _("Missing reply control!\n"));
 		goto error_out;
 	}
 
 	sctrl = ber_init(&pprc->ldctl_value);
 
 	if (!sctrl) {
-		fprintf(stderr, "ber_init() failed, Invalid control ?!\n");
+		fprintf(stderr, _("ber_init() failed, Invalid control ?!\n"));
 		goto error_out;
 	}
 
@@ -612,13 +695,16 @@ static int ldap_set_keytab(krb5_context krbctx,
 
 	rtag = ber_scanf(sctrl, "{i{", &kvno);
 	if (rtag == LBER_ERROR) {
-		fprintf(stderr, "ber_scanf() failed, Invalid control ?!\n");
+		fprintf(stderr, _("ber_scanf() failed, Invalid control ?!\n"));
 		goto error_out;
 	}
 
 	for (i = 0; i < keys->nkeys; i++) {
 		ret = ber_scanf(sctrl, "{i}", &encs[i]);
-		if (ret == LBER_ERROR) break;
+		if (ret == LBER_ERROR) {
+                    fprintf(stderr, _("ber_scanf() failed, Invalid control ?!\n"));
+                    goto error_out;
+                }
 	}
 
 	ret = filter_keys(krbctx, keys, encs);
@@ -629,6 +715,7 @@ static int ldap_set_keytab(krb5_context krbctx,
 	ldap_controls_free(srvctrl);
 	ldap_msgfree(res);
 	ldap_unbind_ext(ld, NULL, NULL);
+	free(encs);
 	return kvno;
 
 error_out:
@@ -653,13 +740,13 @@ static char *ask_password(krb5_context krbctx)
 
     k5d_pw0.length = sizeof(pw0);
     k5d_pw0.data = pw0;
-    ap_prompts[0].prompt = "New Principal Password";
+    ap_prompts[0].prompt = _("New Principal Password");
     ap_prompts[0].hidden = 1;
     ap_prompts[0].reply = &k5d_pw0;
 
     k5d_pw1.length = sizeof(pw1);
     k5d_pw1.data = pw1;
-    ap_prompts[1].prompt = "Verify Principal Password";
+    ap_prompts[1].prompt = _("Verify Principal Password");
     ap_prompts[1].hidden = 1;
     ap_prompts[1].reply = &k5d_pw1;
 
@@ -668,7 +755,7 @@ static char *ask_password(krb5_context krbctx)
                 2, ap_prompts);
 
     if (strcmp(pw0, pw1)) {
-        fprintf(stderr, "Passwords do not match!");
+        fprintf(stderr, _("Passwords do not match!"));
         return NULL;
     }
 
@@ -680,24 +767,43 @@ static char *ask_password(krb5_context krbctx)
     return password;
 }
 
-int main(int argc, char *argv[])
+int main(int argc, const char *argv[])
 {
 	static const char *server = NULL;
 	static const char *principal = NULL;
 	static const char *keytab = NULL;
 	static const char *enctypes_string = NULL;
+	static const char *binddn = NULL;
+	static const char *bindpw = NULL;
 	int quiet = 0;
 	int askpass = 0;
 	int permitted_enctypes = 0;
         struct poptOption options[] = {
-		{ "quiet", 'q', POPT_ARG_NONE, &quiet, 0, "Print as little as possible", "Output only on errors"},
-                { "server", 's', POPT_ARG_STRING, &server, 0, "Contact this specific KDC Server", "Server Name" },
-                { "principal", 'p', POPT_ARG_STRING, &principal, 0, "The principal to get a keytab for (ex: ftp/ftp.example.com@EXAMPLE.COM)", "Kerberos Service Principal Name" },
-                { "keytab", 'k', POPT_ARG_STRING, &keytab, 0, "File were to store the keytab information", "Keytab File Name" },
-		{ "enctypes", 'e', POPT_ARG_STRING, &enctypes_string, 0, "Encryption types to request", "Comma separated encryption types list" },
-		{ "permitted-enctypes", 0, POPT_ARG_NONE, &permitted_enctypes, 0, "Show the list of permitted encryption types and exit", "Permitted Encryption Types"},
-		{ "password", 'P', POPT_ARG_NONE, &askpass, 0, "Asks for a non-random password to use for the principal" },
-		{ NULL, 0, POPT_ARG_NONE, NULL, 0, NULL, NULL }
+            { "quiet", 'q', POPT_ARG_NONE, &quiet, 0,
+              _("Print as little as possible"), _("Output only on errors")},
+            { "server", 's', POPT_ARG_STRING, &server, 0,
+              _("Contact this specific KDC Server"),
+              _("Server Name") },
+            { "principal", 'p', POPT_ARG_STRING, &principal, 0,
+              _("The principal to get a keytab for (ex: ftp/ftp.example.com@EXAMPLE.COM)"),
+              _("Kerberos Service Principal Name") },
+            { "keytab", 'k', POPT_ARG_STRING, &keytab, 0,
+              _("File were to store the keytab information"),
+              _("Keytab File Name") },
+	    { "enctypes", 'e', POPT_ARG_STRING, &enctypes_string, 0,
+              _("Encryption types to request"),
+              _("Comma separated encryption types list") },
+	    { "permitted-enctypes", 0, POPT_ARG_NONE, &permitted_enctypes, 0,
+              _("Show the list of permitted encryption types and exit"),
+              _("Permitted Encryption Types") },
+	    { "password", 'P', POPT_ARG_NONE, &askpass, 0,
+              _("Asks for a non-random password to use for the principal"), NULL },
+	    { "binddn", 'D', POPT_ARG_STRING, &binddn, 0,
+              _("LDAP DN"), _("DN to bind as if not using kerberos") },
+	    { "bindpw", 'w', POPT_ARG_STRING, &bindpw, 0,
+              _("LDAP password"), _("password to use if not using kerberos") },
+            POPT_AUTOHELP
+            POPT_TABLEEND
 	};
 	poptContext pc;
 	char *ktname;
@@ -707,15 +813,19 @@ int main(int argc, char *argv[])
 	krb5_principal uprinc;
 	krb5_principal sprinc;
 	krb5_error_code krberr;
-	ber_int_t *enctypes;
 	struct keys_container keys;
 	krb5_keytab kt;
 	int kvno;
 	int i, ret;
 
+    ret = init_gettext();
+    if (ret) {
+        exit(1);
+    }
+
 	krberr = krb5_init_context(&krbctx);
 	if (krberr) {
-		fprintf(stderr, "Kerberos context initialization failed\n");
+		fprintf(stderr, _("Kerberos context initialization failed\n"));
 		exit(1);
 	}
 
@@ -728,19 +838,20 @@ int main(int argc, char *argv[])
 
 		krberr = krb5_get_permitted_enctypes(krbctx, &ktypes);
 		if (krberr) {
-			fprintf(stderr, "No system preferred enctypes ?!\n");
+			fprintf(stderr, _("No system preferred enctypes ?!\n"));
 			exit(1);
 		}
-		fprintf(stdout, "Supported encryption types:\n");
+		fprintf(stdout, _("Supported encryption types:\n"));
 		for (i = 0; ktypes[i]; i++) {
 			krberr = krb5_enctype_to_string(ktypes[i], enc, 79);
 			if (krberr) {
-				fprintf(stderr, "Warning: failed to convert type (#%d)\n", i);
+				fprintf(stderr, _("Warning: "
+                                        "failed to convert type (#%d)\n"), i);
 				continue;
 			}
 			fprintf(stdout, "%s\n", enc);
 		}
-		krb5_free_ktypes(krbctx, ktypes);
+		ipa_krb5_free_ktypes(krbctx, ktypes);
 		exit (0);
 	}
 
@@ -751,6 +862,14 @@ int main(int argc, char *argv[])
 		exit(2);
 	}
 
+	if (NULL!=binddn && NULL==bindpw) {
+		fprintf(stderr,
+                        _("Bind password required when using a bind DN.\n"));
+		if (!quiet)
+			poptPrintUsage(pc, stderr, 0);
+		exit(10);
+	}
+
         if (askpass) {
 		password = ask_password(krbctx);
 		if (!password) {
@@ -758,7 +877,8 @@ int main(int argc, char *argv[])
 		}
 	} else if (enctypes_string && strchr(enctypes_string, ':')) {
 		if (!quiet) {
-			fprintf(stderr, "Warning: salt types are not honored with randomized passwords (see opt. -P)\n");
+			fprintf(stderr, _("Warning: salt types are not honored"
+                                " with randomized passwords (see opt. -P)\n"));
 		}
 	}
 
@@ -769,38 +889,42 @@ int main(int argc, char *argv[])
 
 	krberr = krb5_parse_name(krbctx, principal, &sprinc);
 	if (krberr) {
-		fprintf(stderr, "Invalid Service Principal Name\n");
+		fprintf(stderr, _("Invalid Service Principal Name\n"));
 		exit(4);
 	}
 
-	krberr = krb5_cc_default(krbctx, &ccache);
-	if (krberr) {
-		fprintf(stderr, "Kerberos Credential Cache not found\n"
-				"Do you have a Kerberos Ticket?\n");
-		exit(5);
-	}
+	if (NULL == bindpw) {
+		krberr = krb5_cc_default(krbctx, &ccache);
+		if (krberr) {
+			fprintf(stderr,
+                                _("Kerberos Credential Cache not found. "
+				  "Do you have a Kerberos Ticket?\n"));
+			exit(5);
+		}
 
-	krberr = krb5_cc_get_principal(krbctx, ccache, &uprinc);
-	if (krberr) {
-		fprintf(stderr, "Kerberos User Principal not found\n"
-				"Do you have a valid Credential Cache?\n");
-		exit(6);
+		krberr = krb5_cc_get_principal(krbctx, ccache, &uprinc);
+		if (krberr) {
+			fprintf(stderr,
+                                _("Kerberos User Principal not found. "
+				  "Do you have a valid Credential Cache?\n"));
+			exit(6);
+		}
 	}
 
 	krberr = krb5_kt_resolve(krbctx, ktname, &kt);
 	if (krberr) {
-		fprintf(stderr, "Failed to open Keytab\n");
+		fprintf(stderr, _("Failed to open Keytab\n"));
 		exit(7);
 	}
 
 	/* create key material */
 	ret = create_keys(krbctx, sprinc, password, enctypes_string, &keys);
 	if (!ret) {
-		fprintf(stderr, "Failed to create key material\n");
+		fprintf(stderr, _("Failed to create key material\n"));
 		exit(8);
 	}
 
-	kvno = ldap_set_keytab(krbctx, server, principal, uprinc, &keys);
+	kvno = ldap_set_keytab(krbctx, server, principal, uprinc, binddn, bindpw, &keys);
 	if (!kvno) {
 		exit(9);
 	}
@@ -814,7 +938,8 @@ int main(int argc, char *argv[])
 
 		krberr = krb5_kt_add_entry(krbctx, kt, &kt_entry);
 		if (krberr) {
-			fprintf(stderr, "Failed to add key to the keytab\n");
+			fprintf(stderr,
+                                _("Failed to add key to the keytab\n"));
 			exit (11);
 		}
 	}
@@ -823,13 +948,13 @@ int main(int argc, char *argv[])
 
 	krberr = krb5_kt_close(krbctx, kt);
 	if (krberr) {
-		fprintf(stderr, "Failed to close the keytab\n");
+		fprintf(stderr, _("Failed to close the keytab\n"));
 		exit (12);
 	}
 
 	if (!quiet) {
 		fprintf(stderr,
-			"Keytab successfully retrieved and stored in: %s\n",
+			_("Keytab successfully retrieved and stored in: %s\n"),
 			keytab);
 	}
 	exit(0);
